'
' Demonstrate a 64 element fixed point FFT routine
' Yielding a 32 element result. i.e. FFT_Elements / 2
'
' The spectrum is displayed on a 16x2 alphanumeric LCD
' The actual FFT routines takes approx 2.8ms to complete when using a 64MHz oscillator
' Making it almost realtime
'
' Written for the Crownhill Proton compiler version 3.5.3.0 onwards using an 18F device
'
' Author Les Johnson 02-02-2012
'
' You are free to use this code as you see fit, however, a mention of the author and compiler would be appreciated.
' The code is offered as-is and Crownhill nor the author take any responsibility for its use.
'
    Include "Amicus18.inc"                  ' Configure for the Amicus18 device (18F25K20) running at 64MHz
    Include "Amicus18_ADC.inc"              ' Load the Amicus18 ADC routines into the program

    $define FFT_Use_Sqr                     ' Instruct the FFT routine to use more refined real and imaginary combining
    $define FFT_Elements 64                 ' The amount of elements for the FFT routine to work on
    Include "FFT.inc"                       ' Load the 8-bit FFT routines into the program
'
' Setup the Alphanumeric LCD
'
    Declare LCD_DTPin = PORTB.4             ' LCD's Data lines (D4 to D7)
    Declare LCD_ENPin = PORTB.3             ' LCD's EN line
    Declare LCD_RSPin = PORTB.2             ' LCD's RS line
    Declare LCD_Interface = 4               ' 4-bit interface to LCD
    Declare LCD_Lines = 2                   ' LCD contains 2 lines
    Declare LCD_Type = Alphanumeric         ' LCD type is alphanumeric
'
' Create variables used within the scale subroutine
'
    Dim Map_bTemp1 As Byte
    Dim Map_bTemp2 As Byte
    Dim Map_wTemp As Word                   ' Holds the result of the scale calculation

    Dim Map_bXin As Map_bTemp1              ' Value is the value to Scale
    Dim Map_bInMin As Byte                  ' Lower bound of the value's current range
    Dim Map_bInMax As Byte                  ' Upper bound of the value's current range
    Dim Map_bOutMin As Byte                 ' Lower bound of the value's target range
    Dim Map_bOutMax As Map_bTemp2           ' Upper bound of the value's target range
'
' Create variables used for display of the spectrum
'
    Dim wHighestOfTwo As Word               ' Holds two elements extracted from the FFT real array
    Dim bIndex As Byte                      ' General purpose index variable
    Dim bLCD_Xpos As Byte                   ' Holds the x position on the LCD
    Dim wCharData As Word                   ' Holds the two individual LCD bar patterns
    Dim bBarValue As Byte                   ' Holds the value for the bar display on the LCD

'--------------------------------------------------------------------------------
    GoTo Main                               ' Jump over the subroutines
'-------------------------------------------------------------------------
' Table of LCD bars required for a given value
'
BarTable:
    CData 32, 32,_
          32, 08, 32, 09, 32, 10, 32, 11, 32, 12, 32, 13, 32, 14, 32, 15,_
          08, 15, 09, 15, 10, 15, 11, 15, 12, 15, 13, 15, 14, 15, 15, 15

'-------------------------------------------------------------------------
' Table of bit patterns for the bars on the LCD
'
CharTable:
    CData $FE, $40,_
          %00000, %00000, %00000, %00000, %00000, %00000, %00000, %11111,_
          %00000, %00000, %00000, %00000, %00000, %00000, %11111, %11111,_
          %00000, %00000, %00000, %00000, %00000, %11111, %11111, %11111,_
          %00000, %00000, %00000, %00000, %11111, %11111, %11111, %11111,_
          %00000, %00000, %00000, %11111, %11111, %11111, %11111, %11111,_
          %00000, %00000, %11111, %11111, %11111, %11111, %11111, %11111,_
          %00000, %11111, %11111, %11111, %11111, %11111, %11111, %11111,_
          %11111, %11111, %11111, %11111, %11111, %11111, %11111, %11111
'-----------------------------------------------------------------------------------------
' Scale one 8-bit value to another 8-bit value
' Input     : Map_bXin holds the value to scale
'           : Map_bInMin is the lower bound of the value's current range
'           : Map_bInMax is the upper bound of the value's current range
'           : Map_bOutMin is the lower bound of the value's target range
'           : Map_bOutMax is the upper bound of the value's target range
' Output    : pResult holds the result
' Notes     : Traps a zero value for extra speed
'
$define Map8(pX, pInMin, pInMax, pOutMin, pOutMax, pResult) '
    Map_bXin = pX         '
    Map_bInMin = pInMin   '
    Map_bInMax = pInMax   '
    Map_bOutMin = pOutMin '
    Map_bOutMax = pOutMax '
    _Map8                 '
    pResult = Map_wTemp

_Map8 Macro-
    GoSub __Map8
Endm

#ifMacro- _Map8
__Map8:
    If Map_bXin = 0 Then                        ' Is the value to scale zero?
        Map_wTemp = 0                           ' Yes. So return a zero
    Else                                        ' Otherwise...
        Map_bTemp1 = Map_bXin - Map_bInMin
        Map_bTemp2 = Map_bOutMax - Map_bOutMin
        Map_wTemp = Map_bTemp1 * Map_bTemp2
        Map_bTemp1 = Map_bInMax - Map_bInMin
        Map_bTemp2 = Map_wTemp / Map_bTemp1
        Map_wTemp = Map_bTemp2 + Map_bOutMin
    EndIf
    Return
#endIfMacro-
'-------------------------------------------------------------------------
' Create the bit patterns that make up the bars in the LCD's CGRAM.
' Input     : Code memory table CharTable holds the bit patterns
' Output    : None
' Notes     : None
'
$define CreateBarPatterns() GoSub _CreateBarPatterns

_CreateBarPatterns:
    bIndex = 0
    Repeat                                              ' Create loop to read from the data table
        Print CRead8 CharTable[bIndex]                  ' Transfer the table values to the LCD's CGRAM
        Inc bIndex                                      ' Move up a byte
    Until bIndex >= 66                                  ' Untill all GCGRAM is filled
    Return
'-------------------------------------------------------------------------
' Display a vertical bar on the alphanumeric LCD
' Input     : bBarValue holds the value to display as a vertical bar on the LCD
'           : bLCD_Xpos is the x position on the LCD to display the bar
' Output    : Prints a bar on the LCD
' Notes     : None
'
$define DisplayBar() GoSub _DisplayBar

_DisplayBar:
    If bBarValue > 0 Then                               ' Is there a bar to display?
        If bBarValue >= 32 Then bBarValue = 32          ' Yes. So make sure a maximum is reached
        Map8(bBarValue, 0,31, 0, 16, bBarValue)         ' Scale the value from (0 to 31) to (0 to 16)
        wCharData = CRead16 BarTable[bBarValue]         ' Read top and bottom bar pattern from the table
        Print At 1, bLCD_Xpos, wCharData.Byte0, At 2, bLCD_Xpos, wCharData.Byte1 ' Display the bar on the LCD
    Else                                                ' Otherwise...
        Print At 1, bLCD_Xpos, " ", At 2, bLCD_Xpos, " " ' Clear the bar
    EndIf
    Return
'-------------------------------------------------------------------------
' The main program loop starts here
'
Main:
'
' Open the ADC for samples taken on AN0
' Left justify for 8-bit result
'
    OpenADC(ADC_FOSC_64 & ADC_LEFT_JUST & ADC_2_TAD, ADC_REF_VDD_VSS, ADC_1ANA)
    Cls                                                     ' Clear the LCD
    CreateBarPatterns()                                     ' Create the bit patterns that make up the LCD's bars

    While 1 = 1                                             ' Create an infinite loop
        '
        ' Fill the real array with 8-bit ADC values
        '
        bIndex = 0
        Repeat
            FFT_bRealData[bIndex] = ReadADC(ADC_CH0)        ' Read the 8-bit ADC value into the array
            DelayUS 52                                      ' Delay for a sample rate of approx 16KHz (FFT range 0 to 8KHz)
            'DelayUS 250                                    ' Delay for a sample rate of approx 4KHz (FFT range 0 to 2KHz)
            Inc bIndex
        Until bIndex >= cFFT_NumberOfSamples                ' Repeat for n samples

        Fix_FFT()                                           ' Perform the FFT (Result in array FFT_bRealData)
'
' Display the spectrum on a 16x2 alphanumeric LCD
' As columns of vertical bars
'
        bLCD_Xpos = 1                                       ' Start at column 1 of the LCD
        bIndex = 0                                          ' Create a loop for all the values in the real array
        Repeat
            '
            ' We only have 16 columns on the LCD, but 32 pieces of data from the FFT, So...
            ' Extract two values from the real array
            ' Find the maximum of the two values and use this as the bar value on the LCD
            '
            wHighestOfTwo.Byte0 = FFT_bRealData[bIndex]     ' Extract the first value
            Inc bIndex
            wHighestOfTwo.Byte1 = FFT_bRealData[bIndex]     ' Extract the second value
            '
            ' Find the highest value of the two and load bBarValue with it
            '
            bBarValue = wHighestOfTwo.Byte0                  ' Default to bBarValue = the first value
            If bBarValue < wHighestOfTwo.Byte1 Then          ' Is bBarValue less than the second value?
                bBarValue = wHighestOfTwo.Byte1              ' Yes. So bBarValue = the second value
            EndIf

            DisplayBar()                                    ' Display a bar on the LCD
            Inc bLCD_Xpos                                   ' Next column on the LCD
            If bLCD_Xpos > 16 Then bLCD_Xpos = 1            ' Back to column 1 when the end of the LCD is reached
            Inc bIndex
        Until bIndex >= cFFT_NumberOfRealElements           ' Until all the elements of the FFT are read
    Wend                                                    ' Do it forever

