'
' Implement a simple tone decoder using a 32 element fixed point FFT routine
' Yielding a 16 element result. i.e. FFT_Elements / 2
'
' The spectrum is displayed on a 16x2 alphanumeric LCD
' And an LED is illuminated for a given frequency range
'
' The actual FFT routines takes approx 1 to complete when using a 64MHz oscillator,
' and the simple combiner for real and imaginary results. Making it almost realtime
'
' Written for the Crownhill Proton compiler version 3.5.3.0 onwards using an 18F device
'
' Author Les Johnson 02-02-2012
'
' You are free to use this code as you see fit, however, a mention of the author and compiler would be appreciated.
' The code is offered as-is and Crownhill nor the author take any responsibility for its use.
'
    Include "Amicus18.inc"                  ' Configure for the Amicus18 device (18F25K20) running at 64MHz
    Include "Amicus18_ADC.inc"              ' Load the Amicus18 ADC routines into the program

    '$define FFT_Use_Sqr                     ' Instruct the FFT routine to use more refined real and imaginary combining
    $define FFT_Elements 32                 ' The amount of elements for the FFT routine to work on
    Include "FFT.inc"                       ' Load the 8-bit FFT routines into the program
'
' Setup the Alphanumeric LCD
'
    Declare LCD_DTPin = PORTB.4             ' LCD's Data lines (D4 to D7)
    Declare LCD_ENPin = PORTB.3             ' LCD's EN line
    Declare LCD_RSPin = PORTB.2             ' LCD's RS line
    Declare LCD_Interface = 4               ' 4-bit interface to LCD
    Declare LCD_Lines = 2                   ' LCD contains 2 lines
    Declare LCD_Type = Alphanumeric         ' LCD type is alphanumeric
'
' Create variables used for display of the spectrum
'
    Dim bIndex As Byte                      ' General purpose index variable
    Dim bBarValue As Byte                   ' Holds the value for the bar display on the LCD
    Dim bLCD_Xpos As Byte System            ' Holds the X position of the LCD and the LED to illuminate or extinguish
    Dim wHighestOfTwo As Word System        ' Holds two elements extracted from the FFT real array

'--------------------------------------------------------------------------------
    GoTo Main                               ' Jump over the subroutines
'-------------------------------------------------------------------------
' Table of bit patterns for the bars on the LCD
'
CharTable:
    CData $FE, $40,_
          %00000, %00000, %00000, %00000, %00000, %00000, %00000, %11111,_
          %00000, %00000, %00000, %00000, %00000, %00000, %11111, %11111,_
          %00000, %00000, %00000, %00000, %00000, %11111, %11111, %11111,_
          %00000, %00000, %00000, %00000, %11111, %11111, %11111, %11111,_
          %00000, %00000, %00000, %11111, %11111, %11111, %11111, %11111,_
          %00000, %00000, %11111, %11111, %11111, %11111, %11111, %11111,_
          %00000, %11111, %11111, %11111, %11111, %11111, %11111, %11111,_
          %11111, %11111, %11111, %11111, %11111, %11111, %11111, %11111

'-------------------------------------------------------------------------
' Create the bit patterns that make up the bars in the LCD's CGRAM.
' Input     : Code memory table CharTable holds the bit patterns
' Output    : None
' Notes     : None
'
$define CreateBarPatterns() GoSub _CreateBarPatterns

_CreateBarPatterns:
    bIndex = 0
    Repeat                                              ' Create loop to read from the data table
        Print CRead8 CharTable[bIndex]                  ' Transfer the table values to the LCD's CGRAM
        Inc bIndex                                      ' Move up a byte
    Until bIndex >= 66                                  ' Untill all GCGRAM is filled
    Return
'-------------------------------------------------------------------------
' Display a vertical bar on the alphanumeric LCD
' And illuminate the correct LED for a given frequency range
' Input     : bBarValue holds the value to display as a vertical bar on the LCD
'           : bLCD_Xpos is the x position on the LCD to display the bar
' Output    : Prints a bar on the LCD
' Notes     : None
'
$define DisplayBar() GoSub _DisplayBar

_DisplayBar:
    If bBarValue > 12 Then                                  ' Is the FFT value over a threshold?
        Print At 1, bLCD_Xpos, $0F, At 2, bLCD_Xpos, $0F    ' Display the bar on the LCD
        SetBit PORTC,bLCD_Xpos - 1
    Else                                                    ' Otherwise...
        Print At 1, bLCD_Xpos, " ", At 2, bLCD_Xpos, " "    ' Clear the bar
        ClearBit PORTC,bLCD_Xpos - 1
    EndIf
    Return
'-------------------------------------------------------------------------
' Find the highest of two values
' Input     : wHighestOfTwo holds the two bytes to measure
' Output    : bBarValue holds the highest of the two bytes
' Notes     :

FindHighest:
    bBarValue = wHighestOfTwo.Byte0                  ' Default to bBarValue = the first value
    If bBarValue < wHighestOfTwo.Byte1 Then          ' Is bBarValue less than the second value?
        bBarValue = wHighestOfTwo.Byte1              ' Yes. So bBarValue = the second value
    EndIf
    Return
'-------------------------------------------------------------------------
' The main program loop starts here
'
Main:
'
' Open the ADC for samples taken on AN0
' Left justify for 8-bit result
'
    OpenADC(ADC_FOSC_64 & ADC_LEFT_JUST & ADC_2_TAD, ADC_REF_VDD_VSS, ADC_1ANA)
    Cls                                                     ' Clear the LCD
    CreateBarPatterns()                                     ' Create the bit patterns that make up the LCD's bars
    Low PORTC                                               ' Extinguish all the LEDs
    While 1 = 1                                             ' Create an infinite loop
        '
        ' Fill the real array with 8-bit ADC values
        '
        bIndex = 0
        Repeat
            FFT_bRealData[bIndex] = ReadADC(ADC_CH0)        ' Read the 8-bit ADC value into the array
            DelayUS 52                                      ' Delay for a sample rate of approx 16KHz (FFT range 0 to 8KHz)
            'DelayUS 250                                    ' Delay for a sample rate of approx 4KHz (FFT range 0 to 2KHz)
            Inc bIndex
        Until bIndex >= cFFT_NumberOfSamples                ' Repeat for n samples

        Fix_FFT()                                           ' Perform the FFT (Result in array FFT_bRealData)
'
' Display the spectrum on a 16x2 alphanumeric LCD as columns of vertical bars
' And illuminate the correct LED according ot the frequency range
'
        bLCD_Xpos = 1                                       ' Start at column 1 of the LCD
        bIndex = 0                                          ' Create a loop for all the values in the real array
        Repeat
            wHighestOfTwo.Byte0 = FFT_bRealData[bIndex]     ' Extract the first value
            Inc bIndex
            wHighestOfTwo.Byte1 = FFT_bRealData[bIndex]     ' Extract the second value
            GoSub FindHighest                               ' Find the highest byte of wHighestOfTwo (result in bBarValue)

            DisplayBar()                                    ' Display a bar on the LCD and illuminate an LED
            Inc bLCD_Xpos                                   ' Next column on the LCD and next LED
            If bLCD_Xpos > 8 Then bLCD_Xpos = 1             ' Back to column 1 when the end of the LCD is reached
            Inc bIndex
        Until bIndex >= cFFT_NumberOfRealElements           ' Until all the elements of the FFT are read
    Wend                                                    ' Do it forever

